@@ -4,7 +4,7 @@
 {-# LANGUAGE TemplateHaskell #-}
 {- |
    Module      : Text.Pandoc.Readers.DocBook
-   Copyright   : Copyright (C) 2006-2023 John MacFarlane
+   Copyright   : Copyright (C) 2006-2024 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -44,8 +44,9 @@ import Text.Pandoc.Builder
 import Text.Pandoc.Class.PandocMonad (PandocMonad, report)
 import Text.Pandoc.Options
 import Text.Pandoc.Logging (LogMessage(..))
-import Text.Pandoc.Shared (safeRead, extractSpaces, headerShift)
+import Text.Pandoc.Shared (safeRead, extractSpaces)
 import Text.Pandoc.Sources (ToSources(..), sourcesToText)
+import Text.Pandoc.Transforms (headerShift)
 import Text.TeXMath (readMathML, writeTeX)
 import qualified Data.Map as M
 import Text.Pandoc.XML.Light
@@ -796,6 +797,9 @@ blockTags = Set.fromList $
 admonitionTags :: [Text]
 admonitionTags = ["caution","danger","important","note","tip","warning"]
 
+titledBlockElements :: [Text]
+titledBlockElements = ["example", "formalpara", "sidebar"]
+
 -- Trim leading and trailing newline characters
 trimNl :: Text -> Text
 trimNl = T.dropAround (== '\n')
@@ -859,16 +863,10 @@ parseBlock (Elem e) =
         "toc"   -> skip -- skip TOC, since in pandoc it's autogenerated
         "index" -> skip -- skip index, since page numbers meaningless
         "para"  -> parseMixed para (elContent e)
-        "formalpara" -> do
-           tit <- case filterChild (named "title") e of
-                        Just t  -> divWith ("",["formalpara-title"],[]) .
-                                   para .  strong <$> getInlines t
-                        Nothing -> return mempty
-           (tit <>) <$> parseMixed para (elContent e)
         "simpara"  -> parseMixed para (elContent e)
         "ackno"  -> parseMixed para (elContent e)
         "epigraph" -> parseBlockquote
-        "blockquote" -> parseBlockquote
+        "blockquote" -> withOptionalTitle parseBlockquote
         "attribution" -> skip
         "titleabbrev" -> skip
         "authorinitials" -> skip
@@ -881,7 +879,8 @@ parseBlock (Elem e) =
         "bibliomisc" -> parseMixed para (elContent e)
         "bibliomixed" -> parseMixed para (elContent e)
         "equation"         -> para <$> equation e displayMath
-        "informalequation" -> para <$> equation e displayMath
+        "informalequation" -> divWith (attrValue "id" e,["informalequation"],[]) .
+                              para <$> equation e displayMath
         "glosssee" -> para . (\ils -> text "See " <> ils <> str ".")
                          <$> getInlines e
         "glossseealso" -> para . (\ils -> text "See also " <> ils <> str ".")
@@ -909,6 +908,7 @@ parseBlock (Elem e) =
         "refsect2" -> sect 2
         "refsect3" -> sect 3
         "refsection" -> gets dbSectionLevel >>= sect . (+1)
+        l | l `elem` titledBlockElements -> parseAdmonition l
         l | l `elem` admonitionTags -> parseAdmonition l
         "area" -> skip
         "areaset" -> skip
@@ -917,9 +917,10 @@ parseBlock (Elem e) =
         "question" -> addToStart (strong (str "Q:") <> str " ") <$> getBlocks e
         "answer" -> addToStart (strong (str "A:") <> str " ") <$> getBlocks e
         "abstract" -> blockQuote <$> getBlocks e
-        "calloutlist" -> bulletList <$> callouts
-        "itemizedlist" -> bulletList . handleCompact <$> listitems
-        "orderedlist" -> do
+        "calloutlist" -> withOptionalTitle $ bulletList <$> callouts
+        "itemizedlist" -> withOptionalTitle $
+                            bulletList . handleCompact <$> listitems
+        "orderedlist" -> withOptionalTitle $ do
           let listStyle = case attrValue "numeration" e of
                                "arabic"     -> Decimal
                                "loweralpha" -> LowerAlpha
@@ -980,6 +981,7 @@ parseBlock (Elem e) =
            lift $ report $ IgnoredElement name
            return mempty
 
+
          compactSpacing = case attrValue "spacing" e of
                             "compact" -> True
                             _         -> False
@@ -1021,6 +1023,11 @@ parseBlock (Elem e) =
                      items' <- mapM getBlocks items
                      return (mconcat $ intersperse (str "; ") terms', items')
          parseTable = do
+                      let elId = attrValue "id" e
+                      let attrs = case attrValue "tabstyle" e of
+                                    "" -> []
+                                    x  -> [("custom-style", x)]
+                      let classes = T.words $ attrValue "class" e
                       let isCaption x = named "title" x || named "caption" x
                       capt <- case filterChild isCaption e of
                                     Just t  -> getInlines t
@@ -1074,7 +1081,8 @@ parseBlock (Elem e) =
                                                 Nothing  -> replicate numrows ColWidthDefault
                       let toRow = Row nullAttr
                           toHeaderRow l = [toRow l | not (null l)]
-                      return $ table (simpleCaption $ plain capt)
+                      return $ tableWith (elId,classes,attrs)
+                                     (simpleCaption $ plain capt)
                                      (zip aligns widths)
                                      (TableHead nullAttr $ toHeaderRow headrows)
                                      [TableBody nullAttr 0 [] $ map toRow bodyrows]
@@ -1099,23 +1107,34 @@ parseBlock (Elem e) =
              Just t  -> Just ("titleabbrev", strContentRecursive t)
              Nothing -> Nothing
          lineItems = mapM getInlines $ filterChildren (named "line") e
+
+         -- <title> elements can be directly nested inside an admonition block, use
+         -- it if it's there. It is unclear whether we should include the label in
+         -- the title: docbook references are ambiguous on that, and some implementations of admonitions
+         -- (e.g. asciidoctor) just use an icon in all cases. To be conservative, we don't
+         -- include the label and leave it to styling.
+         --
+         getTitle = case filterChild (named "title") e of
+                        Just t  -> Just <$> getInlines t
+                        Nothing -> return Nothing
+         withOptionalTitle p = do
+           mbt <- getTitle
+           b <- p
+           case mbt of
+             Nothing -> return b
+             Just t -> return $ divWith (attrValue "id" e,[],[])
+                         (divWith ("", ["title"], []) (plain t) <> b)
+
          -- Admonitions are parsed into a div. Following other Docbook tools that output HTML,
          -- we parse the optional title as a div with the @title@ class, and give the
          -- block itself a class corresponding to the admonition name.
          parseAdmonition label = do
-           -- <title> elements can be directly nested inside an admonition block, use
-           -- it if it's there. It is unclear whether we should include the label in
-           -- the title: docbook references are ambiguous on that, and some implementations of admonitions
-           -- (e.g. asciidoctor) just use an icon in all cases. To be conservative, we don't
-           -- include the label and leave it to styling.
-           title <- divWith ("", ["title"], []) . plain <$>
-                    case filterChild (named "title") e of
-                        Just t  -> getInlines t
-                        Nothing -> return mempty
-           -- this will ignore the title element if it is present
+           mbt <- getTitle
+           -- this will ignore the title element if it is present:
            b <- getBlocks e
+           let t = divWith ("", ["title"], []) (plain $ fromMaybe mempty mbt)
            -- we also attach the label as a class, so it can be styled properly
-           return $ divWith (attrValue "id" e,[label],[]) (title <> b)
+           return $ divWith (attrValue "id" e,[label],[]) (t <> b)
 
 toAlignment :: Element -> Alignment
 toAlignment c = case findAttr (unqual "align") c of
