@@ -7,7 +7,7 @@
 {-# LANGUAGE ScopedTypeVariables   #-}
 {- |
    Module      : Text.Pandoc.Readers.LaTeX.Parsing
-   Copyright   : Copyright (C) 2006-2022 John MacFarlane
+   Copyright   : Copyright (C) 2006-2024 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -34,7 +34,6 @@ module Text.Pandoc.Readers.LaTeX.Parsing
   , getInputTokens
   , untokenize
   , untoken
-  , toksToString
   , satisfyTok
   , peekTok
   , parseFromToks
@@ -118,11 +117,9 @@ import Text.Pandoc.Logging
 import Text.Pandoc.Options
 import Text.Pandoc.Parsing hiding (blankline, many, mathDisplay, mathInline,
                             space, spaces, withRaw, (<|>))
-import Text.Pandoc.Readers.LaTeX.Types (ExpansionPoint (..), Macro (..),
+import Text.Pandoc.TeX (ExpansionPoint (..), Macro (..),
                                         ArgSpec (..), Tok (..), TokType (..))
 import Text.Pandoc.Shared
-import Text.Parsec.Pos
-import Text.Parsec (Stream(uncons))
 import Text.Pandoc.Walk
 
 newtype DottedNum = DottedNum [Int]
@@ -160,7 +157,7 @@ data LaTeXState = LaTeXState{ sOptions       :: ReaderOptions
                             , sLogMessages   :: [LogMessage]
                             , sIdentifiers   :: Set.Set Text
                             , sVerbatimMode  :: Bool
-                            , sCaption       :: Maybe Inlines
+                            , sCaption       :: Maybe Caption
                             , sInListItem    :: Bool
                             , sInTableCell   :: Bool
                             , sLastHeaderNum :: DottedNum
@@ -264,7 +261,7 @@ instance Monad m => Stream TokStream m Tok where
   uncons (TokStream _ []) = return Nothing
   uncons (TokStream _ (t:ts)) = return $ Just (t, TokStream False ts)
 
-type LP m = ParserT TokStream LaTeXState m
+type LP m = ParsecT TokStream LaTeXState m
 
 withVerbatimMode :: PandocMonad m => LP m a -> LP m a
 withVerbatimMode parser = do
@@ -279,7 +276,7 @@ withVerbatimMode parser = do
 
 rawLaTeXParser :: (PandocMonad m, HasMacros s, HasReaderOptions s, Show a)
                => [Tok] -> LP m a -> LP m a
-               -> ParserT Sources s m (a, Text)
+               -> ParsecT Sources s m (a, Text)
 rawLaTeXParser toks parser valParser = do
   pstate <- getState
   let lstate = def{ sOptions = extractReaderOptions pstate }
@@ -319,7 +316,7 @@ rawLaTeXParser toks parser valParser = do
                 return (val, result')
 
 applyMacros :: (PandocMonad m, HasMacros s, HasReaderOptions s)
-            => Text -> ParserT Sources s m Text
+            => Text -> ParsecT Sources s m Text
 applyMacros s = (guardDisabled Ext_latex_macros >> return s) <|>
    do let retokenize = untokenize <$> many anyTok
       pstate <- getState
@@ -347,7 +344,7 @@ tokenizeSources = concatMap tokenizeSource . unSources
 
 -- Return tokens from input sources. Ensure that starting position is
 -- correct.
-getInputTokens :: PandocMonad m => ParserT Sources s m [Tok]
+getInputTokens :: PandocMonad m => ParsecT Sources s m [Tok]
 getInputTokens = do
   pos <- getPosition
   ss <- getInput
@@ -357,41 +354,41 @@ getInputTokens = do
       Sources ((_,t):rest) -> tokenizeSources $ Sources ((pos,t):rest)
 
 tokenize :: SourcePos -> Text -> [Tok]
-tokenize = totoks
+tokenize = totoks False
  where
-  totoks pos t =
+  totoks atIsLetter pos t =
     case T.uncons t of
        Nothing        -> []
        Just (c, rest)
          | c == '\n' ->
            Tok pos Newline "\n"
-           : totoks (setSourceColumn (incSourceLine pos 1) 1) rest
+           : totoks atIsLetter (setSourceColumn (incSourceLine pos 1) 1) rest
          | isSpaceOrTab c ->
            let (sps, rest') = T.span isSpaceOrTab t
            in  Tok pos Spaces sps
-               : totoks (incSourceColumn pos (T.length sps))
+               : totoks atIsLetter (incSourceColumn pos (T.length sps))
                  rest'
          | isAlphaNum c ->
            let (ws, rest') = T.span isAlphaNum t
            in  Tok pos Word ws
-               : totoks (incSourceColumn pos (T.length ws)) rest'
+               : totoks atIsLetter (incSourceColumn pos (T.length ws)) rest'
          | c == '%' ->
            let (cs, rest') = T.break (== '\n') rest
            in  Tok pos Comment ("%" <> cs)
-               : totoks (incSourceColumn pos (1 + T.length cs)) rest'
+               : totoks atIsLetter (incSourceColumn pos (1 + T.length cs)) rest'
          | c == '\\' ->
            case T.uncons rest of
                 Nothing -> [Tok pos (CtrlSeq " ") "\\"]
                 Just (d, rest')
-                  | isLetterOrAt d ->
-                      -- \makeatletter is common in macro defs;
-                      -- ideally we should make tokenization sensitive
-                      -- to \makeatletter and \makeatother, but this is
-                      -- probably best for now
-                      let (ws, rest'') = T.span isLetterOrAt rest
+                  | isLetter' atIsLetter d ->
+                      let (ws, rest'') = T.span (isLetter' atIsLetter) rest
                           (ss, rest''') = T.span isSpaceOrTab rest''
+                          atIsLetter' = case ws of
+                                          "makeatletter" -> True
+                                          "makeatother" -> False
+                                          _ -> atIsLetter
                       in  Tok pos (CtrlSeq ws) ("\\" <> ws <> ss)
-                          : totoks (incSourceColumn pos
+                          : totoks atIsLetter' (incSourceColumn pos
                                (1 + T.length ws + T.length ss)) rest'''
                   | isSpaceOrTab d || d == '\n' ->
                       let (w1, r1) = T.span isSpaceOrTab rest
@@ -404,24 +401,36 @@ tokenize = totoks
                       in  case T.uncons r3 of
                                Just ('\n', _) ->
                                  Tok pos (CtrlSeq " ") ("\\" <> w1)
-                                 : totoks (incSourceColumn pos (T.length ws))
-                                   r1
+                                 : totoks atIsLetter
+                                    (incSourceColumn pos (T.length ws)) r1
                                _ ->
                                  Tok pos (CtrlSeq " ") ws
-                                 : totoks (incSourceColumn pos (T.length ws))
-                                   r3
+                                 : totoks atIsLetter
+                                    (incSourceColumn pos (T.length ws)) r3
                   | otherwise  ->
                       Tok pos (CtrlSeq (T.singleton d)) (T.pack [c,d])
-                      : totoks (incSourceColumn pos 2) rest'
+                      : totoks atIsLetter (incSourceColumn pos 2) rest'
          | c == '#' ->
-           let (t1, t2) = T.span (\d -> d >= '0' && d <= '9') rest
-           in  case safeRead t1 of
-                    Just i ->
-                       Tok pos (Arg i) ("#" <> t1)
-                       : totoks (incSourceColumn pos (1 + T.length t1)) t2
-                    Nothing ->
-                       Tok pos Symbol "#"
-                       : totoks (incSourceColumn pos 1) t2
+           case T.uncons rest of
+             Just ('#', t3) ->
+               let (t1, t2) = T.span (\d -> d >= '0' && d <= '9') t3
+               in  case safeRead t1 of
+                        Just i ->
+                           Tok pos (DeferredArg i) ("##" <> t1)
+                           : totoks atIsLetter
+                              (incSourceColumn pos (2 + T.length t1)) t2
+                        Nothing -> Tok pos Symbol "#"
+                                  : Tok (incSourceColumn pos 1) Symbol "#"
+                                  : totoks atIsLetter (incSourceColumn pos 1) t3
+             _ ->
+               let (t1, t2) = T.span (\d -> d >= '0' && d <= '9') rest
+               in  case safeRead t1 of
+                        Just i ->
+                           Tok pos (Arg i) ("#" <> t1)
+                           : totoks atIsLetter
+                               (incSourceColumn pos (1 + T.length t1)) t2
+                        Nothing -> Tok pos Symbol "#"
+                                  : totoks atIsLetter (incSourceColumn pos 1) rest
          | c == '^' ->
            case T.uncons rest of
                 Just ('^', rest') ->
@@ -431,26 +440,35 @@ tokenize = totoks
                            case T.uncons rest'' of
                                 Just (e, rest''') | isLowerHex e ->
                                   Tok pos Esc2 (T.pack ['^','^',d,e])
-                                  : totoks (incSourceColumn pos 4) rest'''
+                                  : totoks atIsLetter
+                                     (incSourceColumn pos 4) rest'''
                                 _ ->
                                   Tok pos Esc1 (T.pack ['^','^',d])
-                                  : totoks (incSourceColumn pos 3) rest''
+                                  : totoks atIsLetter
+                                      (incSourceColumn pos 3) rest''
                          | d < '\128' ->
                                   Tok pos Esc1 (T.pack ['^','^',d])
-                                  : totoks (incSourceColumn pos 3) rest''
+                                  : totoks atIsLetter
+                                     (incSourceColumn pos 3) rest''
                        _ -> Tok pos Symbol "^" :
                             Tok (incSourceColumn pos 1) Symbol "^" :
-                            totoks (incSourceColumn pos 2) rest'
+                            totoks atIsLetter (incSourceColumn pos 2) rest'
                 _ -> Tok pos Symbol "^"
-                     : totoks (incSourceColumn pos 1) rest
+                     : totoks atIsLetter (incSourceColumn pos 1) rest
          | otherwise ->
-           Tok pos Symbol (T.singleton c) : totoks (incSourceColumn pos 1) rest
+           Tok pos Symbol (T.singleton c) :
+             totoks atIsLetter (incSourceColumn pos 1) rest
 
 isSpaceOrTab :: Char -> Bool
 isSpaceOrTab ' '  = True
 isSpaceOrTab '\t' = True
 isSpaceOrTab _    = False
 
+-- First parameter is True if @ is letter
+isLetter' :: Bool -> Char -> Bool
+isLetter' True '@' = True
+isLetter' _ c = isLetter c
+
 isLetterOrAt :: Char -> Bool
 isLetterOrAt '@' = True
 isLetterOrAt c   = isLetter c
@@ -475,9 +493,6 @@ untokenAccum (Tok _ _ t) accum = t <> accum
 untoken :: Tok -> Text
 untoken t = untokenAccum t mempty
 
-toksToString :: [Tok] -> String
-toksToString = T.unpack . untokenize
-
 parseFromToks :: PandocMonad m => LP m a -> [Tok] -> LP m a
 parseFromToks parser toks = do
   oldInput <- getInput
@@ -486,7 +501,11 @@ parseFromToks parser toks = do
   case toks of
      Tok pos _ _ : _ -> setPosition pos
      _ -> return ()
-  result <- disablingWithRaw parser
+  -- we ignore existing raw tokens maps (see #9517)
+  oldRawTokens <- sRawTokens <$> getState
+  updateState $ \st -> st{ sRawTokens = mempty }
+  result <- parser
+  updateState $ \st -> st{ sRawTokens = oldRawTokens }
   setInput oldInput
   setPosition oldpos
   return result
@@ -571,6 +590,8 @@ doMacros' n inp =
       x <- try $ spaces >> bracedOrToken
       getargs (M.insert i x argmap) rest
 
+    addTok False _args spos (Tok _ (DeferredArg i) txt) acc =
+      Tok spos (Arg i) txt : acc
     addTok False args spos (Tok _ (Arg i) _) acc =
        case M.lookup i args of
             Nothing -> mzero
@@ -864,7 +885,12 @@ dimenarg = try $ do
   optional sp
   ch  <- option False $ True <$ symbol '='
   minus <- option "" $ "-" <$ symbol '-'
-  Tok _ _ s1 <- satisfyTok isWordTok
+  s1 <- option ""
+        (do Tok _ _ s1 <- satisfyTok isWordTok
+            guard (case T.uncons s1 of
+                     Just (c,_) -> isDigit c
+                     Nothing -> False)
+            pure s1)
   s2 <- option "" $ try $ do
           symbol '.'
           Tok _ _ t <-  satisfyTok isWordTok
@@ -872,10 +898,14 @@ dimenarg = try $ do
   let s = s1 <> s2
   let (num, rest) = T.span (\c -> isDigit c || c == '.') s
   guard $ T.length num > 0
-  guard $ rest `elem` ["", "pt","pc","in","bp","cm","mm","dd","cc","sp"]
+  guard $ rest `elem`
+    ["", "pt","pc","in","bp","cm","mm","dd","cc","sp","ex","em",
+     "mu", -- "mu" in math mode only
+     "px" -- "px" with pdftex and luatex only
+    ]
   return $ T.pack ['=' | ch] <> minus <> s
 
-ignore :: (Monoid a, PandocMonad m) => Text -> ParserT s u m a
+ignore :: (Monoid a, PandocMonad m) => Text -> ParsecT s u m a
 ignore raw = do
   pos <- getPosition
   report $ SkippedContent raw pos
@@ -977,6 +1007,18 @@ getRawCommand name txt = do
              void (manyTill anyTok braced) <|>
                 void (satisfyTok isPreTok) -- see #7531
            _ | isFontSizeCommand name -> return ()
+             | name `elem` ["hfil", "hfill", "vfil", "vfill",
+                            "hfilneg", "vfilneg"] -> return ()
+             | name `elem` ["hskip", "vskip", "mskip"] -> do
+                 dimenarg
+                 skipMany $ try $ do
+                   sp
+                   satisfyTok $
+                     \case
+                       Tok _ Word "plus" -> True
+                       Tok _ Word "minus" -> True
+                       _ -> False
+                   dimenarg
              | otherwise -> do
                skipopts
                option "" (try dimenarg)
@@ -1066,10 +1108,11 @@ label = do
 
 setCaption :: PandocMonad m => LP m Inlines -> LP m ()
 setCaption inline = try $ do
-  skipopts
+  mbshort <- Just . toList <$> bracketed inline <|> pure Nothing
   ils <- tokWith inline
   optional $ try $ spaces *> label
-  updateState $ \st -> st{ sCaption = Just ils }
+  updateState $ \st -> st{ sCaption = Just $
+                              Caption mbshort [Plain $ toList ils] }
 
 resetCaption :: PandocMonad m => LP m ()
 resetCaption = updateState $ \st -> st{ sCaption   = Nothing
