@@ -8,7 +8,7 @@
 {-# LANGUAGE OverloadedStrings     #-}
 {- |
    Module      : Text.Pandoc.Shared
-   Copyright   : Copyright (C) 2006-2023 John MacFarlane
+   Copyright   : Copyright (C) 2006-2024 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -50,22 +50,19 @@ module Text.Pandoc.Shared (
                      linesToPara,
                      figureDiv,
                      makeSections,
+                     makeSectionsWithOffsets,
                      combineAttr,
                      uniqueIdent,
                      inlineListToIdentifier,
                      textToIdentifier,
                      isHeaderBlock,
-                     headerShift,
-                     stripEmptyParagraphs,
                      onlySimpleTableCells,
                      isTightList,
                      taskListItemFromAscii,
                      taskListItemToAscii,
                      handleTaskListItem,
                      addMetaField,
-                     eastAsianLineBreakFilter,
                      htmlSpanLikeElements,
-                     filterIpynbOutput,
                      formatCode,
                      -- * TagSoup HTML handling
                      renderTags',
@@ -93,8 +90,7 @@ import Data.Containers.ListUtils (nubOrd)
 import Data.Char (isAlpha, isLower, isSpace, isUpper, toLower, isAlphaNum,
                   generalCategory, GeneralCategory(NonSpacingMark,
                   SpacingCombiningMark, EnclosingMark, ConnectorPunctuation))
-import Data.List (find, foldl', groupBy, intercalate, intersperse,
-                  union, sortOn)
+import Data.List (find, foldl', groupBy, intercalate, intersperse, union)
 import qualified Data.Map as M
 import Data.Maybe (mapMaybe)
 import Data.Monoid (Any (..) )
@@ -102,6 +98,7 @@ import Data.Semigroup (Min (..))
 import Data.Sequence (ViewL (..), ViewR (..), viewl, viewr)
 import qualified Data.Set as Set
 import qualified Data.Text as T
+import qualified Text.Emoji as Emoji
 import System.Directory
 import System.FilePath (isPathSeparator, splitDirectories)
 import qualified System.FilePath.Posix as Posix
@@ -113,7 +110,6 @@ import Data.Time
 import Text.Pandoc.Asciify (toAsciiText)
 import Text.Pandoc.Definition
 import Text.Pandoc.Extensions (Extensions, Extension(..), extensionEnabled)
-import Text.Pandoc.Generic (bottomUp)
 import Text.DocLayout (charWidth)
 import Text.Pandoc.Walk
 -- for addPandocAttributes:
@@ -302,7 +298,7 @@ addPandocAttributes
   :: forall b . HasAttributes (Cm () b) => [(T.Text, T.Text)] -> b -> b
 addPandocAttributes [] bs = bs
 addPandocAttributes kvs bs =
-  unCm . addAttributes (("wrapper","1"):kvs) $ (Cm bs :: Cm () b)
+  unCm . addAttributes kvs $ (Cm bs :: Cm () b)
 
 -- | Generate infinite lazy list of markers for an ordered list,
 -- depending on list attributes.
@@ -469,7 +465,7 @@ isPara _        = False
 -- | Convert Pandoc inline list to plain text identifier.
 inlineListToIdentifier :: Extensions -> [Inline] -> T.Text
 inlineListToIdentifier exts =
-  textToIdentifier exts . stringify . walk unEmojify
+  textToIdentifier exts . stringify . unEmojify
   where
     unEmojify :: [Inline] -> [Inline]
     unEmojify
@@ -477,7 +473,10 @@ inlineListToIdentifier exts =
         extensionEnabled Ext_ascii_identifiers exts = walk unEmoji
       | otherwise = id
     unEmoji (Span ("",["emoji"],[("data-emoji",ename)]) _) = Str ename
+    unEmoji (Str t) = Str (Emoji.replaceEmojis emojisToAliases t)
     unEmoji x = x
+    emojisToAliases t [] = t
+    emojisToAliases _ (a:_) = a
 
 -- | Convert string to plain text identifier.
 textToIdentifier :: Extensions -> T.Text -> T.Text
@@ -513,12 +512,20 @@ textToIdentifier exts =
 -- adjusted so that the lowest header level is n.
 -- (There may still be gaps in header level if the author leaves them.)
 makeSections :: Bool -> Maybe Int -> [Block] -> [Block]
-makeSections numbering mbBaseLevel bs =
-  S.evalState (go bs) []
+makeSections = makeSectionsWithOffsets []
+
+-- | Like 'makeSections', but with a parameter for number offsets
+-- (a list of 'Int's, the first of which is added to the level 1
+-- section number, the second to the level 2, and so on).
+makeSectionsWithOffsets :: [Int] -> Bool -> Maybe Int -> [Block] -> [Block]
+makeSectionsWithOffsets numoffsets numbering mbBaseLevel bs =
+  S.evalState (go bs) numoffsets
  where
   getLevel (Header level _ _) = Min level
   getLevel _ = Min 99
-  minLevel = getMin $ query getLevel bs
+  minLevel = if all (== 0) numoffsets
+                then getMin $ query getLevel bs
+                else 1 -- see #5071, for backwards compatibility
   go :: [Block] -> S.State [Int] [Block]
   go (Header level (ident,classes,kvs) title':xs) = do
     lastnum <- S.get
@@ -600,29 +607,6 @@ isHeaderBlock :: Block -> Bool
 isHeaderBlock Header{} = True
 isHeaderBlock _        = False
 
--- | Shift header levels up or down.
-headerShift :: Int -> Pandoc -> Pandoc
-headerShift n (Pandoc meta (Header m _ ils : bs))
-  | n < 0
-  , m + n == 0 = headerShift n $
-                 B.setTitle (B.fromList ils) $ Pandoc meta bs
-headerShift n (Pandoc meta bs) = Pandoc meta (walk shift bs)
-
- where
-   shift :: Block -> Block
-   shift (Header level attr inner)
-     | level + n > 0  = Header (level + n) attr inner
-     | otherwise      = Para inner
-   shift x            = x
-
--- | Remove empty paragraphs.
-stripEmptyParagraphs :: Pandoc -> Pandoc
-stripEmptyParagraphs = walk go
-  where go :: [Block] -> [Block]
-        go = filter (not . isEmptyParagraph)
-        isEmptyParagraph (Para []) = True
-        isEmptyParagraph _         = False
-
 -- | Detect if table rows contain only cells consisting of a single
 -- paragraph that has no @LineBreak@.
 onlySimpleTableCells :: [[[Block]]] -> Bool
@@ -692,64 +676,10 @@ addMetaField key val (Meta meta) =
         tolist (MetaList ys) = ys
         tolist y             = [y]
 
--- | Remove soft breaks between East Asian characters.
-eastAsianLineBreakFilter :: Pandoc -> Pandoc
-eastAsianLineBreakFilter = bottomUp go
-  where go (x:SoftBreak:y:zs)
-          | Just (_, b) <- T.unsnoc $ stringify x
-          , Just (c, _) <- T.uncons $ stringify y
-          , charWidth b == 2
-          , charWidth c == 2
-          = x:y:zs
-          | otherwise
-          = x:SoftBreak:y:zs
-        go xs
-          = xs
-
 -- | Set of HTML elements that are represented as Span with a class equal as
 -- the element tag itself.
 htmlSpanLikeElements :: Set.Set T.Text
-htmlSpanLikeElements = Set.fromList ["kbd", "mark", "dfn"]
-
--- | Process ipynb output cells.  If mode is Nothing,
--- remove all output.  If mode is Just format, select
--- best output for the format.  If format is not ipynb,
--- strip out ANSI escape sequences from CodeBlocks (see #5633).
-filterIpynbOutput :: Maybe Format -> Pandoc -> Pandoc
-filterIpynbOutput mode = walk go
-  where go (Div (ident, "output":os, kvs) bs) =
-          case mode of
-            Nothing  -> Div (ident, "output":os, kvs) []
-            -- "best" for ipynb includes all formats:
-            Just fmt
-              | fmt == Format "ipynb"
-                          -> Div (ident, "output":os, kvs) bs
-              | otherwise -> Div (ident, "output":os, kvs) $
-                              walk removeANSI $
-                              take 1 $ sortOn rank bs
-                 where
-                  rank (RawBlock (Format "html") _)
-                    | fmt == Format "html" = 1 :: Int
-                    | fmt == Format "markdown" = 3
-                    | otherwise = 4
-                  rank (RawBlock (Format "latex") _)
-                    | fmt == Format "latex" = 1
-                    | fmt == Format "markdown" = 3
-                    | otherwise = 4
-                  rank (RawBlock f _)
-                    | fmt == f = 1
-                    | otherwise = 4
-                  rank (Para [Image{}]) = 2
-                  rank _ = 3
-                  removeANSI (CodeBlock attr code) =
-                    CodeBlock attr (removeANSIEscapes code)
-                  removeANSI x = x
-                  removeANSIEscapes t
-                    | Just cs <- T.stripPrefix "\x1b[" t =
-                        removeANSIEscapes $ T.drop 1 $ T.dropWhile (/='m') cs
-                    | Just (c, cs) <- T.uncons t = T.cons c $ removeANSIEscapes cs
-                    | otherwise = ""
-        go x = x
+htmlSpanLikeElements = Set.fromList ["kbd", "mark", "dfn", "abbr"]
 
 -- | Reformat 'Inlines' as code, putting the stringlike parts in 'Code'
 -- elements while bringing other inline formatting outside.
