@@ -18,9 +18,13 @@ module Text.Pandoc.Writers.Typst (
   ) where
 import Text.Pandoc.Definition
 import Text.Pandoc.Class ( PandocMonad)
-import Text.Pandoc.Options ( WriterOptions(..), WrapOption(..), isEnabled )
+import Text.Pandoc.ImageSize ( dimension, Dimension(Pixel), Direction(..),
+                               showInInch )
+import Text.Pandoc.Options ( WriterOptions(..), WrapOption(..), isEnabled,
+                             CaptionPosition(..) )
 import Data.Text (Text)
 import Data.List (intercalate, intersperse)
+import Data.Bifunctor (first, second)
 import Network.URI (unEscapeString)
 import qualified Data.Text as T
 import Control.Monad.State ( StateT, evalStateT, gets, modify )
@@ -36,6 +40,7 @@ import Text.Pandoc.Extensions (Extension(..))
 import Text.Collate.Lang (Lang(..), parseLang)
 import Text.Printf (printf)
 import Data.Char (isAlphaNum)
+import Data.Maybe (fromMaybe)
 
 -- | Convert Pandoc to Typst.
 writeTypst :: PandocMonad m => WriterOptions -> Pandoc -> m Text
@@ -65,6 +70,9 @@ pandocToTypst options (Pandoc meta blocks) = do
               (fmap chomp . inlinesToTypst)
               meta
   main <- blocksToTypst blocks
+  let toPosition :: CaptionPosition -> Text
+      toPosition CaptionAbove = "top"
+      toPosition CaptionBelow = "bottom"
   let context = defField "body" main
               $ defField "toc" (writerTableOfContents options)
               $ (if isEnabled Ext_citations options
@@ -78,7 +86,13 @@ pandocToTypst options (Pandoc meta blocks) = do
                         Right l ->
                           resetField "lang" (langLanguage l) .
                           maybe id (resetField "region") (langRegion l))
+              $ defField "smart" (isEnabled Ext_smart options)
               $ defField "toc-depth" (tshow $ writerTOCDepth options)
+              $ defField "figure-caption-position"
+                   (toPosition $ writerFigureCaptionPosition options)
+              $ defField "table-caption-position"
+                   (toPosition $ writerTableCaptionPosition options)
+              $ defField "page-numbering" ("1" :: Text)
               $ (if writerNumberSections options
                     then defField "numbering" ("1.1.1.1.1" :: Text)
                     else id)
@@ -88,6 +102,45 @@ pandocToTypst options (Pandoc meta blocks) = do
        Nothing  -> main
        Just tpl -> renderTemplate tpl context
 
+pickTypstAttrs :: [(Text, Text)] -> ([(Text, Text)],[(Text, Text)])
+pickTypstAttrs = foldr go ([],[])
+  where
+    go (k,v) =
+      case T.splitOn ":" k of
+        "typst":"text":x:[] -> second ((x,v):)
+        "typst":x:[] -> first ((x,v):)
+        _ -> id
+
+formatTypstProp :: (Text, Text) -> Text
+formatTypstProp (k,v) = k <> ": " <> v
+
+toTypstPropsListSep :: [(Text, Text)] -> Doc Text
+toTypstPropsListSep = hsep . intersperse "," . (map $ literal . formatTypstProp)
+
+toTypstPropsListTerm :: [(Text, Text)] -> Doc Text
+toTypstPropsListTerm [] = ""
+toTypstPropsListTerm typstAttrs = toTypstPropsListSep typstAttrs <> ","
+
+toTypstPropsListParens :: [(Text, Text)] -> Doc Text
+toTypstPropsListParens [] = ""
+toTypstPropsListParens typstAttrs = parens $ toTypstPropsListSep typstAttrs
+
+toTypstTextElement :: [(Text, Text)] -> Doc Text -> Doc Text
+toTypstTextElement [] content = content
+toTypstTextElement typstTextAttrs content = "#text" <> toTypstPropsListParens typstTextAttrs <> brackets content
+
+toTypstSetText :: [(Text, Text)] -> Doc Text
+toTypstSetText [] = ""
+toTypstSetText typstTextAttrs = "set text" <> parens (toTypstPropsListSep typstTextAttrs) <> "; "
+
+toTypstPoundSetText :: [(Text, Text)] -> Doc Text
+toTypstPoundSetText [] = ""
+toTypstPoundSetText typstTextAttrs = "#" <> toTypstSetText typstTextAttrs
+
+toTypstBracesSetText :: [(Text, Text)] -> Doc Text -> Doc Text
+toTypstBracesSetText [] x = "#" <> x
+toTypstBracesSetText typstTextAttrs x = "#" <> braces (toTypstSetText typstTextAttrs <> x)
+
 blocksToTypst :: PandocMonad m => [Block] -> TW m (Doc Text)
 blocksToTypst blocks = vcat <$> mapM blockToTypst blocks
 
@@ -144,8 +197,9 @@ blockToTypst block =
                                   parens (
                                     "numbering: " <>
                                     doubleQuoted
-                                      (head (orderedListMarkers
-                                             (1, sty, delim))) <>
+                                     (case orderedListMarkers (1, sty, delim) of
+                                          (m:_) -> m
+                                          [] -> "1.") <>
                                     ", start: " <>
                                       text (show start) )) $$
                                x $$
@@ -163,13 +217,14 @@ blockToTypst block =
                    else vsep items') $$ blankline
     DefinitionList items ->
       ($$ blankline) . vsep <$> mapM defListItemToTypst items
-    Table (ident,_,_) (Caption _ caption) colspecs thead tbodies tfoot -> do
+    Table (ident,tabclasses,tabkvs) (Caption _ caption) colspecs thead tbodies tfoot -> do
       let lab = toLabel FreestandingLabel ident
       capt' <- if null caption
                   then return mempty
                   else do
                     captcontents <- blocksToTypst caption
                     return $ ", caption: " <> brackets captcontents
+      let typstFigureKind = literal (", kind: " <> fromMaybe "table" (lookup "typst:figure:kind" tabkvs))
       let numcols = length colspecs
       let (aligns, widths) = unzip colspecs
       let commaSep = hcat . intersperse ", "
@@ -185,26 +240,40 @@ blockToTypst block =
           formatalign AlignCenter = "center,"
           formatalign AlignDefault = "auto,"
       let alignarray = parens $ mconcat $ map formatalign aligns
-      let fromCell (Cell _attr alignment rowspan colspan bs) = do
-            let cellattrs =
+
+      let fromCell (Cell (_,_,kvs) alignment rowspan colspan bs) = do
+            let (typstAttrs, typstTextAttrs) = pickTypstAttrs kvs
+            let valign =
+                  (case lookup "align" typstAttrs of
+                    Just va -> [va]
+                    _ -> [])
+            let typstAttrs2 = filter ((/="align") . fst) typstAttrs
+            let halign =
                   (case alignment of
-                     AlignDefault -> []
-                     AlignLeft -> [ "align: left" ]
-                     AlignRight -> [ "align: right" ]
-                     AlignCenter -> [ "align: center" ]) ++
+                    AlignDefault -> []
+                    AlignLeft -> [ "left" ]
+                    AlignRight -> [ "right" ]
+                    AlignCenter -> [ "center" ])
+            let cellaligns = valign ++ halign
+            let cellattrs =
+                  (case cellaligns of
+                    [] -> []
+                    _ -> [ "align: " <> T.intercalate " + " cellaligns ]) ++
                   (case rowspan of
                      RowSpan 1 -> []
                      RowSpan n -> [ "rowspan: " <> tshow n ]) ++
                   (case colspan of
                      ColSpan 1 -> []
-                     ColSpan n -> [ "colspan: " <> tshow n ])
+                     ColSpan n -> [ "colspan: " <> tshow n ]) ++
+                  map formatTypstProp typstAttrs2
             cellContents <- blocksToTypst bs
+            let contents2 = brackets (toTypstPoundSetText typstTextAttrs <> cellContents)
             pure $ if null cellattrs
-                      then brackets cellContents
+                      then contents2
                       else "table.cell" <>
                             parens
                              (literal (T.intercalate ", " cellattrs)) <>
-                            brackets cellContents
+                            contents2
       let fromRow (Row _ cs) =
             (<> ",") . commaSep <$> mapM fromCell cs
       let fromHead (TableHead _attr headRows) =
@@ -223,35 +292,40 @@ blockToTypst block =
             hrows <- mapM fromRow headRows
             brows <- mapM fromRow bodyRows
             pure $ vcat (hrows ++ ["table.hline()," | not (null hrows)] ++ brows)
+      let (typstAttrs, typstTextAttrs) = pickTypstAttrs tabkvs
       header <- fromHead thead
       footer <- fromFoot tfoot
       body <- vcat <$> mapM fromTableBody tbodies
-      return $
-        "#figure("
-        $$
-        nest 2
-         ("align(center)[#table("
-          $$ nest 2
-             (  "columns: " <> columns <> ","
-             $$ "align: " <> alignarray <> ","
-             $$ header
-             $$ body
-             $$ footer
-             )
-          $$ ")]"
-          $$ capt'
-          $$ ", kind: table"
-          $$ ")")
-        $$ lab
-        $$ blankline
+      let table = "table("
+            $$ nest 2
+                (  "columns: " <> columns <> ","
+                $$ "align: " <> alignarray <> ","
+                $$ toTypstPropsListTerm typstAttrs
+                $$ header
+                $$ body
+                $$ footer
+            )
+            $$ ")"
+      return $ if "typst:no-figure" `elem` tabclasses
+        then toTypstBracesSetText typstTextAttrs table
+        else "#figure("
+            $$
+            nest 2
+            ("align(center)[" <> toTypstPoundSetText typstTextAttrs <> "#" <> table <> "]"
+              $$ capt'
+              $$ typstFigureKind
+              $$ ")")
+            $$ lab
+          $$ blankline
     Figure (ident,_,_) (Caption _mbshort capt) blocks -> do
       caption <- blocksToTypst capt
+      opts <-  gets stOptions
       contents <- case blocks of
                      -- don't need #box around block-level image
-                     [Para [Image (_,_,kvs) _ (src, _)]]
-                       -> pure $ mkImage False src kvs
-                     [Plain [Image (_,_,kvs) _ (src, _)]]
-                       -> pure $ mkImage False src kvs
+                     [Para [Image attr _ (src, _)]]
+                       -> pure $ mkImage opts False src attr
+                     [Plain [Image attr _ (src, _)]]
+                       -> pure $ mkImage opts False src attr
                      _ -> brackets <$> blocksToTypst blocks
       let lab = toLabel FreestandingLabel ident
       return $ "#figure(" <> nest 2 ((contents <> ",")
@@ -261,10 +335,13 @@ blockToTypst block =
                           $$ ")" $$ lab $$ blankline
     Div (ident,_,_) (Header lev ("",cls,kvs) ils:rest) ->
       blocksToTypst (Header lev (ident,cls,kvs) ils:rest)
-    Div (ident,_,_) blocks -> do
+    Div (ident,_,kvs) blocks -> do
       let lab = toLabel FreestandingLabel ident
+      let (typstAttrs,typstTextAttrs) = pickTypstAttrs kvs
       contents <- blocksToTypst blocks
-      return $ "#block[" $$ contents $$ ("]" <+> lab)
+      return $ "#block" <> toTypstPropsListParens typstAttrs <> "["
+        $$ toTypstPoundSetText typstTextAttrs <> contents
+        $$ ("]" <+> lab)
 
 defListItemToTypst :: PandocMonad m => ([Inline], [[Block]]) -> TW m (Doc Text)
 defListItemToTypst (term, defns) = do
@@ -273,7 +350,7 @@ defListItemToTypst (term, defns) = do
   modify $ \st -> st{ stEscapeContext = NormalContext }
   defns' <- mapM blocksToTypst defns
   return $ nowrap ("/ " <> term' <> ": " <> "#block[") $$
-            chomp (vcat defns') $$ "]"
+            chomp (vsep defns') $$ "]"
 
 listItemToTypst :: PandocMonad m => Int -> Doc Text -> [Block] -> TW m (Doc Text)
 listItemToTypst ind marker blocks = do
@@ -287,8 +364,9 @@ inlineToTypst :: PandocMonad m => Inline -> TW m (Doc Text)
 inlineToTypst inline =
   case inline of
     Str txt -> do
+      opts <-  gets stOptions
       context <- gets stEscapeContext
-      return $ escapeTypst context txt
+      return $ escapeTypst (isEnabled Ext_smart opts) context txt
     Space -> return space
     SoftBreak -> do
       wrapText <- gets $ writerWrapText . stOptions
@@ -323,15 +401,26 @@ inlineToTypst inline =
     Superscript inlines -> textstyle "#super" inlines
     Subscript inlines -> textstyle "#sub" inlines
     SmallCaps inlines -> textstyle "#smallcaps" inlines
-    Span (ident,_,_) inlines -> do
+    Span (ident,_,kvs) inlines -> do
       let lab = toLabel FreestandingLabel ident
-      (<> lab) <$> inlinesToTypst inlines
+      let (_, typstTextAttrs) = pickTypstAttrs kvs
+      case typstTextAttrs of
+        [] -> (<> lab) <$> inlinesToTypst inlines
+        _ -> do
+          contents <- inlinesToTypst inlines
+          return $ toTypstTextElement typstTextAttrs contents <> lab
     Quoted quoteType inlines -> do
-      let q = case quoteType of
-                   DoubleQuote -> literal "\""
-                   SingleQuote -> literal "'"
+      opts <- gets stOptions
+      let smart = isEnabled Ext_smart opts
       contents <- inlinesToTypst inlines
-      return $ q <> contents <> q
+      return $
+        case quoteType of
+           DoubleQuote
+             | smart -> "\"" <> contents <> "\""
+             | otherwise -> "“" <> contents <> "”"
+           SingleQuote
+             | smart -> "'" <> contents <> "'"
+             | otherwise -> "‘" <> contents <> "’"
     Cite citations inlines -> do
       opts <-  gets stOptions
       if isEnabled Ext_citations opts
@@ -355,24 +444,34 @@ inlineToTypst inline =
                     (if inlines == [Str src]
                           then mempty
                           else nowrap $ brackets contents) <> endCode
-    Image (_,_,kvs) _inlines (src,_tit) -> pure $ mkImage True src kvs
+    Image attr _inlines (src,_tit) -> do
+      opts <-  gets stOptions
+      pure $ mkImage opts True src attr
     Note blocks -> do
       contents <- blocksToTypst blocks
       return $ "#footnote" <> brackets (chomp contents) <> endCode
 
 -- see #9104; need box or image is treated as block-level
-mkImage :: Bool -> Text -> [(Text, Text)] -> Doc Text
-mkImage useBox src kvs
+mkImage :: WriterOptions -> Bool -> Text -> Attr -> Doc Text
+mkImage opts useBox src attr
   | useBox = "#box" <> parens coreImage
   | otherwise = coreImage
  where
   src' = T.pack $ unEscapeString $ T.unpack src -- #9389
-  toDimAttr k =
-     case lookup k kvs of
-       Just v -> ", " <> literal k <> ": " <> literal v
-       Nothing -> mempty
-  dimAttrs = mconcat $ map toDimAttr ["height", "width"]
-  coreImage = "image" <> parens (doubleQuoted src' <> dimAttrs)
+  showDim (Pixel a) = literal (showInInch opts (Pixel a) <> "in")
+  showDim dim = text (show dim)
+  dimAttrs =
+     (case dimension Height attr of
+        Nothing -> mempty
+        Just dim -> ", height: " <> showDim dim) <>
+     (case dimension Width attr of
+        Nothing -> mempty
+        Just dim -> ", width: " <> showDim dim)
+  isData = "data:" `T.isPrefixOf` src'
+  dataSvg = "<svg xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\"><image xlink:href=\"" <> src' <> "\" /></svg>"
+  coreImage
+    | isData = "image.decode" <> parens(doubleQuoted dataSvg <> dimAttrs)
+    | otherwise = "image" <> parens (doubleQuoted src' <> dimAttrs)
 
 textstyle :: PandocMonad m => Doc Text -> [Inline] -> TW m (Doc Text)
 textstyle s inlines =
@@ -386,8 +485,8 @@ textstyle s inlines =
          , needsEscapeAtLineStart c -> ("\\" <>)
        _ -> id
 
-escapeTypst :: EscapeContext -> Text -> Doc Text
-escapeTypst context t =
+escapeTypst :: Bool -> EscapeContext -> Text -> Doc Text
+escapeTypst smart context t =
   (case T.uncons t of
     Just (c, _)
       | needsEscapeAtLineStart c
@@ -400,9 +499,17 @@ escapeTypst context t =
   where
     escapeChar c
       | c == '\160' = "~"
+      | c == '\8217', smart = "'" -- apostrophe
+      | c == '\8212', smart = "---" -- em dash
+      | c == '\8211', smart = "--" -- en dash
       | needsEscape c = "\\" <> T.singleton c
       | otherwise = T.singleton c
     needsEscape '\160' = True
+    needsEscape '\8217' = smart
+    needsEscape '\8212' = smart
+    needsEscape '\8211' = smart
+    needsEscape '\'' = smart
+    needsEscape '"' = smart
     needsEscape '[' = True
     needsEscape ']' = True
     needsEscape '#' = True
@@ -411,8 +518,6 @@ escapeTypst context t =
     needsEscape '@' = True
     needsEscape '$' = True
     needsEscape '\\' = True
-    needsEscape '\'' = True
-    needsEscape '"' = True
     needsEscape '`' = True
     needsEscape '_' = True
     needsEscape '*' = True
@@ -457,8 +562,7 @@ toCite cite = do
      then do
        suppl <- case citationSuffix cite of
                   [] -> pure mempty
-                  suff -> (<> endCode) . brackets
-                            <$> inlinesToTypst (eatComma suff)
+                  suff -> brackets <$> inlinesToTypst (eatComma suff)
        pure $ "@" <> literal ident' <> suppl
      else do
        let label = if T.all isIdentChar ident'
