@@ -235,6 +235,14 @@ tests = [ testGroup "document"
             "collapse overlapping targets (anchor spans)"
             "docx/overlapping_targets.docx"
             "docx/overlapping_targets.native"
+          , testCompare
+            "anchor in header after anchor"
+            "docx/anchor_header_after_anchor.docx"
+            "docx/anchor_header_after_anchor.native"
+          , testCompare
+            "text in shape format"
+            "docx/text_in_shape_format.docx"
+            "docx/text_in_shape_format.native"
           ]
         , testGroup "blocks"
           [ testCompare
@@ -293,6 +301,10 @@ tests = [ testGroup "document"
             "definition lists"
             "docx/definition_list.docx"
             "docx/definition_list.native"
+          , testCompare
+            "task lists"
+            "docx/task_list.docx"
+            "docx/task_list.native"
           , testCompare
             "custom defined lists in styles"
             "docx/german_styled_lists.docx"
@@ -316,7 +328,7 @@ tests = [ testGroup "document"
           , testCompare
             "blockquotes (parsing indent as blockquote)"
             "docx/block_quotes.docx"
-            "docx/block_quotes_parse_indent.native"
+            "docx/block_quotes.native"
           , testCompare
             "blockquotes (parsing indent relative to the indent of the parent style as blockquote)"
             "docx/relative_indentation_blockquotes.docx"
@@ -475,6 +487,11 @@ tests = [ testGroup "document"
             "comment warnings (all)"
             "docx/comments_warning.docx"
             ["Docx comment 1 will not retain formatting"]
+          , testForWarningsWithOpts def{readerTrackChanges=AllChanges,
+                                        readerExtensions=extensionsFromList [Ext_styles]}
+            "comments (with styles extension)"
+            "docx/comments.docx"
+            []
           ]
         , testGroup "media"
           [ testMediaBag
